﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace tfm.ExeStrings
{
    class Program
    {
        // Code for the ReadFully function taken from http://www.yoda.arachsys.com/csharp/readbinary.html

        /// <summary>
        /// Reads data from a stream until the end is reached. The
        /// data is returned as a byte array. An IOException is
        /// thrown if any of the underlying IO calls fail.
        /// </summary>
        /// <param name="stream">The stream to read data from</param>
        /// <param name="initialLength">The initial buffer length</param>
        public static byte[] ReadFully(Stream stream, int initialLength)
        {
            // If we've been passed an unhelpful initial length, just
            // use 32K.
            if (initialLength < 1)
            {
                initialLength = 32768;
            }

            byte[] buffer = new byte[initialLength];
            int read = 0;

            int chunk;
            while ((chunk = stream.Read(buffer, read, buffer.Length - read)) > 0)
            {
                read += chunk;

                // If we've reached the end of our buffer, check to see if there's
                // any more information
                if (read == buffer.Length)
                {
                    int nextByte = stream.ReadByte();

                    // End of stream? If so, we're done
                    if (nextByte == -1)
                    {
                        return buffer;
                    }

                    // Nope. Resize the buffer, put in the byte we've just
                    // read, and continue
                    byte[] newBuffer = new byte[buffer.Length * 2];
                    Array.Copy(buffer, newBuffer, buffer.Length);
                    newBuffer[read] = (byte)nextByte;
                    buffer = newBuffer;
                    read++;
                }
            }
            // Buffer is now too big. Shrink it.
            byte[] ret = new byte[read];
            Array.Copy(buffer, ret, read);
            return ret;
        }

        /// <summary>
        /// Returns true if the byte represents a digit, a letter (A-Z, a-z) or the underscore.
        /// </summary>
        public static bool Readable(byte b)
        {
            return (b > 47 && b < 58) || (b > 64 && b < 91) || b == 95 || (b > 96 && b < 123);
        }

        static void Main(string[] args)
        {
            if (!File.Exists(args[0]))
            {
                Console.WriteLine("File \"" + args[0] + "\" not found!");
                return;
            }
            byte[] result;
            using (FileStream fs = new FileStream(args[0], FileMode.Open))
            {
                int len = Convert.ToInt32(fs.Length);
                Console.WriteLine("Reading \"" + args[0] + "\" (" + (len / 1024).ToString() + " KB)...");
                result = ReadFully(fs, len);
            }
            List<byte> word = new List<byte>(); // This will contain each word while it's composed
            List<string> words = new List<string>(); // This will contain all the composed words
            bool write = false;
            for (int i=0; i<result.Length; i++)
            {
                byte b = result[i];
                if (Readable(b))
                    write = true;
                else
                {
                    // If we find a 0 between 2 readable characters we discard it,
                    // but we don't count it as the end of a word since most strings
                    // have 2-bytes characters and the first one is 0.
                    if (write &&
                        !(i > 0 && i < result.Length - 1 && b == 0 && Readable(result[i - 1]) && Readable(result[i + 1])))
                    {
                        write = false;
                        if (word.Count > 2) // The word will be added only if it's more than 2 characters
                            words.Add(Encoding.ASCII.GetString(word.ToArray()));
                        word.Clear();
                    }
                }
                if (write && b != 0)
                    word.Add(b);
            }
            Console.WriteLine("Done! Produced " + words.Count.ToString() + " lines.");
            File.WriteAllLines(Path.ChangeExtension(args[0], "txt"), words.ToArray());
        }
    }
}
